# {{crate}}
## Current version: {{version}}

{{badges}}
[![crates.io](https://img.shields.io/crates/v/{{crate}}.svg)](https://crates.io/crates/{{crate}})
[![docs.rs](https://docs.rs/{{crate}}/badge.svg)](https://docs.rs/{{crate}}/)
[![dependency status](https://deps.rs/crate/{{crate}}/{{version}}/status.svg)](https://deps.rs/crate/{{crate}}/{{version}})

{{readme}}

## License : {{license}}
